from typing import Any, Iterable, TypeAlias

from dataclasses import dataclass


@dataclass
class Variable:
    type: str | None = None
    value: Any | None = None
    description: str | None = None
    required: bool = False


Section: TypeAlias = dict[str, Variable]
Config: TypeAlias = dict[str, Section]


FileLike: TypeAlias = Iterable[str] | str



__all__ = ["read_config", "read_configs"]

from .types import Config, FileLike, Variable

from dataclasses import dataclass
import dataclasses

import ast
import re


@dataclass
class Regexes:
    section: re.Pattern[str]
    comment: re.Pattern[str]
    variable: re.Pattern[str]


regexes = Regexes(
    section=re.compile(r"^\[(?P<section>.*)\](#.*)?$"),
    comment=re.compile(r"^[#;]\s?(?P<comment>.*)$"),
    variable=re.compile(
        r"^(?P<name>\w+)\s*(:\s*(?P<type>\w+))?\s*((?P<not_required>=)\s*(?P<value>.*))?\s*([#;].*)?$"
    ),
)


def read_config(config: FileLike, preconfig: Config | None = None) -> Config:
    if isinstance(config, str):
        with open(config) as file:
            return read_config(file, preconfig)
    result: Config = dict(**(preconfig or {}))
    result.setdefault("DEFAULT", {})

    @dataclass
    class Current:
        section: str
        description: list[str]

    current = Current(section="DEFAULT", description=[])
    for i in config:
        match i.strip():
            case "":
                current.description = []

            case i if m := regexes.section.match(i):
                current.section = m.group("section")
                result[current.section] = result.get(current.section, {})
                current.description = []

            case i if m := regexes.comment.match(i):
                current.description += [m.group("comment")]

            case i if m := regexes.variable.match(i):
                variable_name = m.group("name").strip()
                variable_value = ast.literal_eval(m.group("value") or "None")
                variable_exact_type = m.group("type")
                variable_inferred_type = variable_exact_type or (
                    typename
                    if (typename := type(variable_value).__name__) != "NoneType"
                    else None
                )
                current_variable = result[current.section].get(
                    variable_name, Variable()
                )

                result[current.section][variable_name] = dataclasses.replace(
                    current_variable,
                    value=variable_value,
                    required=m.group("not_required") is None,
                    **(
                        (
                            dict(description="\n".join(current.description))
                            if current.description
                            else {}
                        )
                        | (
                            dict(type=variable_inferred_type)
                            if (current_variable.type is None or variable_exact_type)
                            and variable_inferred_type
                            else {}
                        )
                    ),
                )

                current.description = []

            case i:
                raise ValueError(f"Could not parse line: {i}")

    return result


def read_configs(
    filenames: FileLike | list[FileLike], preconfig: Config | None = None
) -> Config:
    result = dict(**(preconfig or {}))
    if not isinstance(filenames, list):
        filenames = [filenames]

    for file in filenames:
        result = read_config(file, result)

    return result



__all__ = ["schema_from_inis"]

from typing import Any, Protocol
from click.decorators import FC
from .types import FileLike

import builtins

import click
from .config import read_configs


class Decorator(Protocol):
    def __call__(self, func: FC) -> FC:
        ...


def schema_from_inis(
    files: list[FileLike] | FileLike = ["config.default.ini", "config.ini"],
    insecure_eval: bool = False,
    **kwargs: Any,
) -> Decorator:
    """Decorate a click command to load options from a config file.

    Parameters
    ----------
    filenames : list[FileLike] | files, optional
        List of files (either open file-pointers or filenames) to load, by default ["config.default.ini", "config.ini"]
    insecure_eval : bool, optional
        Whether or not to allow arbitrary code execution, by default False
    **kwargs : Any
        Passed to click.option
    """

    if isinstance(files, str):
        files = [files]

    config = read_configs(files)

    def decorator(func: FC) -> FC:
        # We use reverse-order so that the options are added in the order they appear in the file
        for section, options in reversed(config.items()):
            for variable_name, d in reversed(options.items()):
                option_name = (
                    f"{f'{section}.' if section != 'DEFAULT' else ''}{variable_name}"
                )

                type_evalled = d.type
                if type_evalled is not None:
                    if insecure_eval:
                        try:
                            type_evalled = eval(type_evalled)
                        except Exception:
                            type_evalled = None
                    else:
                        type_evalled = getattr(builtins, type_evalled, None)

                func = click.option(
                    f"--{option_name}"
                    if type_evalled is not bool
                    else f"--{option_name}/--no-{option_name}",
                    option_name.replace(".", "__"),
                    **(
                        dict(
                            type=type_evalled,
                            default=d.value,
                            required=d.required,
                            help=f"\n{d.description or ''}".replace("\n", "\b\n"),
                            show_default=True,
                        )
                        | kwargs
                    ),
                )(func)

        return func

    return decorator
